'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fs = require('fs');
var path = require('path');
var serveStatic = require('sirv');
var serializeJavascript = require('serialize-javascript');
var querystring = require('querystring');
var router = require('@egoist/router');
var createDebug = require('debug');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var serveStatic__default = /*#__PURE__*/_interopDefaultLegacy(serveStatic);
var serializeJavascript__default = /*#__PURE__*/_interopDefaultLegacy(serializeJavascript);
var createDebug__default = /*#__PURE__*/_interopDefaultLegacy(createDebug);

const TYPE = "content-type";
const OSTREAM = "application/octet-stream";
function send(res, data = "") {
  let type = res.getHeader("content-type");
  if (data && typeof data.pipe === "function") {
    res.setHeader(TYPE, type || OSTREAM);
    return data.pipe(res);
  }
  if (data instanceof Buffer) {
    type = type || OSTREAM;
  } else if (typeof data === "object") {
    data = JSON.stringify(data);
    type = type || "application/json;charset=utf-8";
  } else {
    type = type || "text/html";
  }
  res.setHeader("content-length", Buffer.byteLength(data));
  res.setHeader("content-type", type);
  res.end(data);
}
function status(res, code) {
  res.statusCode = code;
}

const parseUrl = (url) => {
  const queryIndex = url.indexOf("?");
  if (queryIndex === -1) {
    return {path: url, search: ""};
  }
  const path = url.substring(0, queryIndex);
  const search = url.substring(queryIndex);
  return {path, search};
};
class ConnectError extends Error {
}
const onError = (error, _, res) => {
  const status = typeof error === "string" ? 500 : error.status || 500;
  res.statusCode = status;
  res.end(`${status}`);
};
const notFoundError = new ConnectError("404 not found");
notFoundError.status = 404;
class Connect {
  constructor(options = {}) {
    this.get = this.add.bind(this, "GET");
    this.head = this.add.bind(this, "HEAD");
    this.patch = this.add.bind(this, "PATCH");
    this.options = this.add.bind(this, "OPTIONS");
    this.connect = this.add.bind(this, "CONNECT");
    this.delete = this.add.bind(this, "DELETE");
    this.trace = this.add.bind(this, "TRACE");
    this.post = this.add.bind(this, "POST");
    this.put = this.add.bind(this, "PUT");
    this.handler = (_req, _res) => {
      const req = _req;
      const res = _res;
      const info = parseUrl(req.url);
      req.originalUrl = req.originalUrl || req.url;
      req.path = info.path;
      req.query = querystring.parse(info.search.substring(1));
      const matches = this.router.find(req.method, req.url);
      matches.push({handler: this.onNoMatch, params: {}});
      let i = 0;
      const next = (error) => {
        if (error)
          return this.onError(error, req, res);
        const m = matches[i++];
        if (m) {
          req.params = m.params;
          m.handler(req, res, next);
        }
      };
      next();
    };
    this.router = new router.Router();
    this.onError = options.onError || onError;
    this.onNoMatch = options.onNoMatch || this.onError.bind(null, notFoundError);
  }
  add(method, route, ...handlers) {
    this.router.add(method, route, ...handlers);
    return this;
  }
  use(route, ...handlers) {
    if (typeof route === "string") {
      this.router.use(route, ...handlers);
    } else if (typeof route === "function") {
      this.router.use("/*", route, ...handlers);
    }
    return this;
  }
}
const connect = (options) => new Connect(options);

const createServer = (options) => {
  const app = connect(options);
  app.use((req, res, next) => {
    Object.defineProperties(res, {
      send: {
        enumerable: true,
        value: function(data) {
          send(res, data);
          return res;
        }
      },
      status: {
        enumerable: true,
        value: function(statusCode) {
          status(res, statusCode);
          return res;
        }
      }
    });
    next();
  });
  return app;
};

const debug = {
  request: createDebug__default['default']("ream:request"),
  server: createDebug__default['default']("ream:server")
};

const defaultApiHandler = async (server, {serverExports}) => {
  for (const route of serverExports.apiRoutes) {
    server.use(route.path, await route.load().then((res) => res.default));
  }
};
const defaultAppHandler = async (server, {serverExports, htmlTemplatePath, transformIndexHTML, initialState}) => {
  debug.request("called default app handler");
  const htmlTemplate = await fs__default['default'].promises.readFile(htmlTemplatePath, "utf8");
  const MAIN_PLACEHOLDER = `<!--ream-main-->`;
  const HEAD_PLACEHOLDER = `<!--ream-head-->`;
  const HTML_ATTRS_PLACEHOLDER = ` ream-html-attrs`;
  const BODY_ATTRS_PLACEHOLDER = ` ream-body-attrs`;
  server.use(async (req, res) => {
    let html = htmlTemplate;
    if (transformIndexHTML) {
      html = await transformIndexHTML(req.url, html);
    }
    const result = await serverExports.serverRender({
      url: req.url,
      routes: [],
      initialState
    });
    html = html.replace(MAIN_PLACEHOLDER, `
    ${MAIN_PLACEHOLDER}
    <script>INITIAL_STATE=${serializeJavascript__default['default'](initialState, {
      isJSON: true
    })}</script>`);
    if (result) {
      html = html.replace(HTML_ATTRS_PLACEHOLDER, result.htmlAttrs || "").replace(BODY_ATTRS_PLACEHOLDER, result.bodyAttrs || "").replace(HEAD_PLACEHOLDER, result.head || "").replace(MAIN_PLACEHOLDER, result.html || '<div id="_ream"></div>');
    }
    debug.request("sending html");
    res.send(html);
  });
};
async function createHandler(options) {
  const cwd = path__default['default'].resolve(options.cwd || ".");
  const dotReamDir = path__default['default'].join(cwd, ".ream");
  const server = createServer();
  let serverExports;
  const htmlTemplatePath = options.dev ? path__default['default'].join(cwd, "index.html") : path__default['default'].join(dotReamDir, "client/index.html");
  if (options.viteServer) {
    serverExports = await options.viteServer.ssrLoadModule("/.ream/generated/server-exports.js").then((res) => {
      return res;
    });
  } else {
    serverExports = require(path__default['default'].join(dotReamDir, "server/server-exports"));
  }
  if (options.devMiddleware) {
    server.use(options.devMiddleware);
  }
  if (!options.dev) {
    const serveStaticFiles = serveStatic__default['default'](path__default['default'].join(dotReamDir, "client"));
    server.use((req, res, next) => {
      if (req.path === "/")
        return next();
      return serveStaticFiles(req, res, next);
    });
  }
  await serverExports.enhanceServer.callAsync("extendServer", {
    server
  });
  const apiHandler = serverExports.serverHandler ? serverExports.serverHandler.default : defaultApiHandler;
  const appHandler = serverExports.appHandler ? serverExports.appHandler.default : defaultAppHandler;
  const context = {
    serverExports,
    htmlTemplatePath,
    transformIndexHTML: options.viteServer?.transformIndexHtml.bind(options.viteServer),
    initialState: {}
  };
  await apiHandler(server, context);
  await appHandler(server, context);
  return {handler: server.handler};
}

exports.Connect = Connect;
exports.createHandler = createHandler;
