import path from 'path';
import fs from 'fs-extra';

const reamAliasPlugin = (api) => {
  return {
    name: `ream:alias`,
    resolveId(source) {
      return void 0;
    }
  };
};
const reamForceServerUpdatePlugin = (api) => {
  return {
    name: `ream:force-server-update`,
    handleHotUpdate(ctx) {
      const {moduleGraph} = api.viteServer;
      for (const [key, value] of moduleGraph.fileToModulesMap.entries()) {
        if (!key.includes("node_modules")) {
          for (const mod of value) {
            moduleGraph.invalidateModule(mod);
          }
        }
      }
    }
  };
};
const moveManifestPlugin = (manifestDir) => {
  return {
    name: "ream:move-manifest",
    async writeBundle(options, bundle) {
      const files = [];
      for (const file in bundle) {
        const value = bundle[file];
        if (file === "manifest.json" || file === "ssr-manifest.json" && value.type === "asset") {
          files.push({
            name: file === "manifest.json" ? "client-manifest.json" : file,
            originalPath: path.join(options.dir, file),
            content: value.source
          });
          delete bundle[file];
        }
      }
      if (files.length > 0) {
        await Promise.all(files.map(async (file) => {
          await fs.remove(file.originalPath);
          await fs.outputFile(path.join(manifestDir, file.name), file.content, "utf8");
        }));
      }
    }
  };
};
const getViteConfig = (api, server) => {
  const ssrManifest = !server && !api.isDev;
  const entry = api.isDev ? void 0 : server ? api.store.state.apiRoutes.reduce((res, route) => {
    return {
      ...res,
      ["api/" + route.name]: route.file
    };
  }, {"server-exports": api.resolveDotReam("generated/server-exports.js")}) : {
    client: "./index.html"
  };
  const config = {
    mode: api.mode,
    root: api.rootDir,
    configFile: false,
    plugins: [
      reamAliasPlugin(),
      reamForceServerUpdatePlugin(api),
      moveManifestPlugin(api.resolveDotReam("manifest"))
    ],
    define: api.constants,
    resolve: {
      alias: {
        "@": api.resolveSrcDir(),
        "dot-ream": api.resolveDotReam()
      }
    },
    optimizeDeps: {
      exclude: [
        "@ream/fetch",
        "node-fetch",
        "@vue/server-renderer",
        "@ream/framework-vue"
      ]
    },
    ssr: {},
    server: {
      middlewareMode: true,
      hmr: api.config.hmr && {
        host: api.config.hmr.host,
        port: api.config.hmr.port
      }
    },
    build: {
      minify: !api.isDev && !process.env.DEBUG,
      outDir: server ? api.resolveDotReam("server") : api.resolveDotReam("client"),
      ssrManifest,
      ssr: server,
      manifest: !api.isDev && !server,
      rollupOptions: entry ? {
        input: entry
      } : {}
    }
  };
  for (const plugin of api.plugins) {
    if (plugin.plugin.vite) {
      plugin.plugin.vite.call(plugin.context, config);
    }
  }
  if (api.config.vite) {
    api.config.vite(config, {dev: api.isDev});
  }
  return config;
};

export { getViteConfig };
