import path, { resolve, dirname, relative } from 'path';
import { loadEnv } from 'vite';
import resolveFrom from 'resolve-from';
import consola from 'consola';
import JoyCon from 'joycon';
import fs, { pathExists as pathExists$1, remove } from 'fs-extra';
import { Module } from 'module';
import { transformSync } from 'esbuild';
import createDebug from 'debug';
import glob from 'fast-glob';
import { comparePathParserScore, createParser } from '@egoist/router';
import fs$1 from 'fs';

const joycon = new JoyCon();
const transpileAndRequire = (code, filepath, isTS) => {
  const result = transformSync(code, {
    sourcefile: filepath,
    loader: isTS ? "ts" : "js",
    target: "es2019",
    format: "cjs"
  });
  const m = requireFromString(result.code, filepath);
  return m.default || m;
};
joycon.addLoader({
  test: /\.[jt]s$/,
  async load(filepath) {
    const code = await fs.readFile(filepath, "utf8");
    return transpileAndRequire(code, filepath, filepath.endsWith(".ts"));
  }
});
const SUPPORTED_CONFIG_FILES = ["ream.config.js", "ream.config.ts"];
function loadConfig(cwd) {
  joycon.clearCache();
  return joycon.load(SUPPORTED_CONFIG_FILES, cwd);
}
function requireFromString(code, filename, opts) {
  opts = opts || {};
  const appendPaths = opts.appendPaths || [];
  const prependPaths = opts.prependPaths || [];
  if (typeof code !== "string") {
    throw new Error("code must be a string, not " + typeof code);
  }
  const paths = Module._nodeModulePaths(path.dirname(filename));
  const parent = module.parent || void 0;
  const m = new Module(filename, parent);
  m.filename = filename;
  m.paths = [...prependPaths, ...paths, ...appendPaths];
  m._compile(code, filename);
  const exports = m.exports;
  parent && parent.children && parent.children.splice(parent.children.indexOf(m), 1);
  return exports;
}

const debug = createDebug("ream:routes-plugin");
const filenameToRoutePath = (filename, isApi = false) => {
  let routePath = "/" + filename.replace(/^index$/, "").replace(/\[([^\]+])\]/g, (_, p1) => {
    if (p1.startsWith("...")) {
      return `:${p1.substring(3)}(.+)`;
    }
    return `:${p1}`;
  });
  if (isApi) {
    routePath = routePath.replace(/\/index$/, "");
  }
  return routePath;
};
const filepathToRoute = (dir, filepath, isApi = false) => {
  const filename = path.relative(dir, filepath).replace(/\.[a-z]+$/, "");
  const routePath = filenameToRoutePath(filename, isApi);
  return {
    id: filepath,
    name: filename,
    path: routePath,
    file: filepath
  };
};
const routesPlugin = () => {
  const pageExtRe = /\.(jsx|tsx|js|ts|vue|svelte)$/;
  const apiExtRe = /\.[jt]s$/;
  let pagesDir;
  let apiDir;
  let pageFiles;
  let apiFiles;
  return {
    name: `ream:routes`,
    async getClientRoutes() {
      pagesDir = this.resolveSrcDir("pages");
      pageFiles = pageFiles || await glob("**/*.{jsx,tsx,js,ts,vue,svelte}", {
        cwd: pagesDir
      });
      return pageFiles.map((file) => {
        const absolutePath = path.join(pagesDir, file);
        return filepathToRoute(pagesDir, absolutePath);
      });
    },
    async getApiRoutes() {
      apiDir = this.resolveSrcDir("api");
      apiFiles = apiFiles || await glob("**/*.{js,ts}", {
        cwd: apiDir
      });
      return apiFiles.map((file) => {
        const absolutePath = path.join(apiDir, file);
        return filepathToRoute(apiDir, absolutePath, true);
      });
    },
    async onFileChange(event, filepath) {
      if (event !== "add" && event !== "unlink")
        return;
      let isApi = false;
      let isPage = false;
      if (filepath.startsWith(pagesDir) && pageExtRe.test(filepath)) {
        isPage = true;
      }
      if (filepath.startsWith(apiDir) && apiExtRe.test(filepath)) {
        isApi = true;
      }
      if (!isApi && !isPage)
        return;
      const files = isApi ? apiFiles : pageFiles;
      if (!files)
        return;
      debug("New route, isApi: %s, isPage: $s", isApi, isPage);
      const filename = path.relative(pagesDir, filepath);
      if (filename.startsWith("_") || filename.includes("/_"))
        return;
      if (event === "add") {
        files.push(filename);
      } else if (event === "unlink") {
        files.splice(files.indexOf(filename), 1);
      }
      if (isPage) {
        await this.store.writeCommonExports();
      } else {
        await this.store.writeServerExports();
      }
    }
  };
};

const writeFile = (filepath, content, encoding) => {
  fs$1.mkdirSync(path.dirname(filepath), {recursive: true});
  return fs$1.writeFileSync(filepath, content, encoding);
};
const pathExists = (filepath) => new Promise((resolve) => {
  fs$1.access(filepath, (err) => resolve(!err));
});
const writeFileIfChanged = (filepath, content) => {
  if (fs$1.existsSync(filepath)) {
    const prev = fs$1.readFileSync(filepath, "utf8");
    if (prev === content)
      return;
  }
  writeFile(filepath, content, "utf8");
};

const resolveFile = async (files, dir = ".") => {
  for (const file of files) {
    const absolutePath = path.resolve(dir, file);
    if (await pathExists$1(absolutePath)) {
      return absolutePath;
    }
  }
};

const makeNestedRoutes = (_routes) => {
  const routes = [];
  for (const route of _routes) {
    const pathParts = route.path === "/" ? ["index"] : route.path.substring(1).split("/");
    route.path = "";
    let parent = routes;
    route._nest_key_ = route._nest_key_ || "";
    for (let i = 0; i < pathParts.length; i++) {
      let part = pathParts[i];
      const prevPart = pathParts[i - 1];
      const nextPart = pathParts[i + 1];
      const expectedParentKey = route._nest_key_;
      route._nest_key_ += route._nest_key_ ? `/${part}` : part;
      const child = parent.find((parentRoute) => parentRoute._nest_key_ === expectedParentKey);
      if (child) {
        child.children = child.children || [];
        parent = child.children;
        if (part === "index") {
          route.path = "";
        } else {
          route.path = part.replace(/^\//, "");
        }
      } else if (part === "index") {
        if (nextPart) {
          route.path += "/index";
        } else if (prevPart) {
          route.path += "";
        } else if (!prevPart) {
          route.path += "/";
        }
      } else {
        route.path += `/${part}`;
      }
    }
    parent.push(route);
  }
  return {routes};
};
const stringifyClientRoutes = (routes, generatedDir) => {
  const stringifyRoute = (route) => `{
    name: "${route.name}",
    path: "${route.path}",
    load: () => import("${path.relative(generatedDir, route.file)}"),
    ${route.children && route.children.length > 0 ? `children: ${stringifyClientRoutes(route.children, generatedDir)}` : ""}
  }`;
  return `[
    ${routes.map((route) => stringifyRoute(route)).join(",\n")}
  ]`;
};

const getInitialState = () => {
  const state = {
    constants: {},
    appHandler: void 0,
    apiHandler: void 0,
    apiRoutes: [],
    pluginsFiles: {
      app: new Set(),
      server: new Set()
    }
  };
  return state;
};
class Store {
  constructor(api) {
    this.api = api;
    this.state = getInitialState();
  }
  resetState() {
    this.state = getInitialState();
  }
  defineConstant(name, value) {
    this.state.constants[name] = JSON.stringify(value);
  }
  addPluginFile(type, file) {
    this.state.pluginsFiles[type].add(normalizePath(file));
  }
  async writeIndexHtml() {
    const filepath = this.api.resolveSrcDir("index.html");
    if (await pathExists(filepath))
      return;
    consola.info(`Writing ${filepath}`);
    writeFileIfChanged(filepath, `<!DOCTYPE html>
      <html ream-html-attrs>
        <head>
          <meta charset="UTF-8" />
          <meta http-equiv="X-UA-Compatible" content="IE=edge" />
          <meta name="viewport" content="width=device-width, initial-scale=1.0" />
          <!--ream-head-->
        </head>
        <body ream-body-attrs>
          <!--ream-main-->
          <script type="module" src="/.ream/generated/client-entry.js"></script>
        </body>
      </html>
      `);
  }
  async writeCommonExports() {
    let clientRoutes = [];
    for (const plugin of this.api.plugins) {
      if (plugin.plugin.getClientRoutes) {
        const routes = await plugin.plugin.getClientRoutes.call(plugin.context);
        clientRoutes = [...clientRoutes, ...routes];
      }
    }
    if (this.api.config.clientRoutes) {
      clientRoutes = await this.api.config.clientRoutes(clientRoutes);
    }
    const {routes: nestedClientRoutes} = makeNestedRoutes(clientRoutes);
    let content = `
    export var clientRoutes = ${stringifyClientRoutes(nestedClientRoutes, this.api.resolveDotReam("generated"))}
    `;
    const filepath = this.api.resolveDotReam("generated/common-exports.js");
    for (const plugin of this.api.plugins) {
      if (plugin.plugin.transformCommonExports) {
        content = await plugin.plugin.transformCommonExports.call(plugin.context, content);
      }
    }
    writeFileIfChanged(filepath, content);
  }
  async writeGlobalImports() {
    writeFileIfChanged(this.api.resolveDotReam("generated/global-imports.js"), `
      ${this.api.config.imports.map((file) => `import ${JSON.stringify(file)}`).join("\n")}
      `);
  }
  async writeClientEntry() {
    writeFileIfChanged(this.api.resolveDotReam("generated/client-entry.js"), `import * as main from '@/main'
      import {clientRoutes} from './common-exports.js'

      const context = {
        initialState: window.INITIAL_STATE,
        routes: clientRoutes, 
        get url() {
          return location.pathname + location.search
        }
      }

      main.default && main.default(context)
      `);
  }
  async writeServerExports() {
    let apiRoutes = [];
    for (const plugin of this.api.plugins) {
      if (plugin.plugin.getApiRoutes) {
        const routes = await plugin.plugin.getApiRoutes.call(plugin.context);
        apiRoutes = [...apiRoutes, ...routes];
      }
    }
    if (this.api.config.apiRoutes) {
      consola.info(`Loading user provided api routes`);
      apiRoutes = await this.api.config.apiRoutes(apiRoutes);
    }
    this.state.apiRoutes = apiRoutes;
    let content = `
    export var apiRoutes = [
      ${apiRoutes.sort((a, b) => {
      return comparePathParserScore(createParser(a.path), createParser(b.path));
    }).map((route) => {
      return `{
          path: "${route.path}",
          load: () => import("${route.file}")
        }`;
    }).join(",\n")}
    ]

    import * as enhanceServer from './ream.server.js'
    
    export { enhanceServer }
    `;
    const {apiHandler, appHandler} = this.state;
    if (apiHandler) {
      content += `
      import apiHandler from '${normalizePath(apiHandler)}'
      export { apiHandler }
      `;
    }
    if (appHandler) {
      content += `
      import appHandler from '${normalizePath(appHandler)}'
      export { appHandler }
      `;
    }
    content += `
    export const serverRender = async (renderContext) => {
      if (!REAM_SSR_ENABLED) return

      const main = await import('@/main')
      const { clientRoutes } = await import('./common-exports.js') 
      renderContext.routes = clientRoutes
      
      const result = await main.default && main.default(renderContext) 

      return result
    }
    `;
    for (const plugin of this.api.plugins) {
      if (plugin.plugin.transformServerExports) {
        content = await plugin.plugin.transformServerExports.call(plugin.context, content);
      }
    }
    writeFileIfChanged(this.api.resolveDotReam("generated/server-exports.js"), content);
  }
  async writeHookFile(type, projectFiles) {
    const {pluginsFiles} = this.api.store.state;
    const files = [...pluginsFiles[type]];
    const projectFile = await resolveFile(projectFiles);
    if (projectFile) {
      files.push(projectFile);
    }
    let content = `
    ${files.map((file, index) => {
      return `import * as hook_${type}_${index} from "${file}"`;
    }).join("\n")}
  
  var files = [
    ${files.map((_, i) => `hook_${type}_${i}`).join(",")}
  ]
  
  function getExportByName(name) {
    var fns = []
    for (var i = 0; i < files.length; i++) {
      var mod = files[i]
      if (mod[name]) {
        fns.push(mod[name])
      }
    }
    return fns
  }
  
  export async function callAsync(name, context) {
    for (const fn of getExportByName(name)) {
      await fn(context)
    }
  }
  `;
    if (type === "server") {
      content += `
    export async function getInitialHTML(context) {
      let html
      for (const fn of getExportByName('getInitialHTML')) {
        const result = fn(context)
        if (result) {
          html = result
        }
      }
      return html
    }
  
    export const hasExport = (name) => getExportByName(name).length > 0
    `;
    }
    writeFileIfChanged(this.api.resolveDotReam(`generated/ream.${type}.js`), content);
  }
}

class PluginContext {
  constructor(api, pluginName) {
    this.api = api;
    this.pluginName = pluginName;
  }
  get dev() {
    return this.api.isDev;
  }
  get store() {
    return this.api.store;
  }
  get env() {
    return this.api.env;
  }
  get constants() {
    return this.api.constants;
  }
  get config() {
    return this.api.config;
  }
  ensureEnv(name, defaultValue) {
    const value = this.env[name];
    if (!value && !defaultValue) {
      throw new Error(`Plugin "${this.pluginName}" requires environment variable "${name}" to be set.`);
    }
    return value || defaultValue;
  }
  resolveSrcDir(...args) {
    return this.api.resolveSrcDir(...args);
  }
  resolveRootDir(...args) {
    return this.api.resolveRootDir(...args);
  }
  resolveDotReam(...args) {
    return this.api.resolveDotReam(...args);
  }
}

function normalizePath(path) {
  return path.replace(/\\/g, "/");
}

const defineConfig = (config) => config;
class Ream {
  constructor(options = {}, inlineConfig = {}) {
    this.inlineConfig = inlineConfig;
    this.rootDir = resolve(options.rootDir || ".");
    this.isDev = Boolean(options.dev);
    this.mode = options.mode || (this.isDev ? "development" : "production");
    this.store = new Store(this);
    this.plugins = [];
    if (!process.env.NODE_ENV) {
      process.env.NODE_ENV = this.isDev ? "development" : "production";
    }
  }
  resolveRootDir(...args) {
    return resolve(this.rootDir, ...args);
  }
  resolveSrcDir(...args) {
    if (!this.config.srcDir)
      throw new Error(`srcDir is only available after loading config file`);
    return this.resolveRootDir(this.config.srcDir, ...args);
  }
  resolveDotReam(...args) {
    return this.resolveRootDir(".ream", ...args);
  }
  resolveInPackage(pkg, target) {
    const pkgDir = dirname(resolveFrom(process.cwd(), `${pkg}/package.json`));
    return resolveFrom(pkgDir, target);
  }
  async loadConfig() {
    const {data: projectConfig = {}, path: configPath} = await loadConfig(this.rootDir);
    if (configPath) {
      consola.info(`Using config file: ${relative(process.cwd(), configPath)}`);
    }
    this.configPath = configPath;
    this.config = {
      srcDir: ".",
      ...this.inlineConfig,
      ...projectConfig,
      env: {
        ...projectConfig.env,
        ...this.inlineConfig.env
      },
      plugins: [
        routesPlugin(),
        ...this.inlineConfig.plugins || [],
        ...projectConfig.plugins || []
      ],
      imports: projectConfig.imports || []
    };
  }
  get env() {
    return {
      ...this.userEnv,
      ...this.config.env,
      REAM_SOURCE_DIR: this.resolveSrcDir(),
      REAM_ROOT_DIR: this.resolveRootDir()
    };
  }
  get constants() {
    const {env} = this;
    return {
      ...this.store.state.constants,
      ...Object.keys(env).reduce((res, key) => {
        const value = JSON.stringify(env[key]);
        return {
          ...res,
          [`import.meta.env.${key}`]: value,
          [`process.env.${key}`]: value
        };
      }, {}),
      REAM_SSR_ENABLED: JSON.stringify(this.config.ssr !== false)
    };
  }
  getPlugins() {
    return this.config.plugins.map((plugin) => {
      const context = new PluginContext(this, plugin.name);
      return {
        plugin,
        context
      };
    });
  }
  async prepare() {
    this.store.resetState();
    await this.loadConfig();
    this.userEnv = loadEnv(this.mode, this.rootDir, "REAM_");
    this.plugins = this.getPlugins();
    await Promise.all(["generated", "server", "client"].map((name) => {
      return remove(this.resolveDotReam(name));
    }));
    consola.info("Preparing Ream files");
    const {prepareFiles} = await import('./prepare-files-f3a990d7.mjs');
    await prepareFiles(this);
    if (this.isDev) {
      const {createServer: createViteServer} = await import('vite');
      const {getViteConfig} = await import('./get-vite-config-cc12cca2.mjs');
      const viteConfig = getViteConfig(this);
      const viteServer = await createViteServer(viteConfig);
      if (this.viteServer) {
        for (const key in viteServer) {
          if (key !== "app") {
            this.viteServer[key] = viteServer[key];
          }
        }
      } else {
        this.viteServer = viteServer;
      }
      for (const plugin of this.plugins) {
        if (plugin.plugin.onFileChange) {
          viteServer.watcher.on("all", plugin.plugin.onFileChange.bind(plugin.context));
        }
      }
      viteServer.watcher.on("all", async (_, file) => {
        const files = SUPPORTED_CONFIG_FILES.map((name) => this.resolveRootDir(name));
        if (file === this.configPath || files.includes(file)) {
          consola.info(`Restarting Ream due to changes in ${file}`);
          await viteServer.close();
          await this.prepare();
        }
      });
    }
  }
  localResolve(name) {
    return resolveFrom.silent(this.rootDir, name);
  }
  localRequire(name) {
    const path = this.localResolve(name);
    return path && require(path);
  }
  async getRequestHandler() {
    if (this.isDev) {
      await this.prepare();
      const {getRequestHandler} = await import('./dev-server-b13342eb.mjs');
      return getRequestHandler(this);
    }
    const {createHandler} = await import('./server.mjs');
    const {handler} = await createHandler({
      dev: this.isDev,
      cwd: this.rootDir
    });
    return handler;
  }
  async build({
    fullyExport,
    standalone
  }) {
    await this.prepare();
    const {build, buildStandalone} = await import('./build-d15d98a0.mjs');
    await build(this);
    const {exportSite} = await import('./export-d560f1d6.mjs');
    await exportSite(this.resolveDotReam(), fullyExport);
    if (standalone) {
      await buildStandalone(this);
    }
  }
}

export { Ream, defineConfig, normalizePath };
