import * as http from 'http';
import { InlineConfig, UserConfig, ViteDevServer } from 'vite';

/** Mark some properties as required, leaving others unchanged */
declare type MarkRequired<T, RK extends keyof T> = Exclude<T, RK> & Required<Pick<T, RK>>;

declare type State = {
    constants: {
        [k: string]: string;
    };
    appHandler?: string;
    apiHandler?: string;
    apiRoutes: Route[];
    pluginsFiles: {
        app: Set<string>;
        server: Set<string>;
    };
};
declare class Store {
    private api;
    state: State;
    constructor(api: Ream);
    resetState(): void;
    defineConstant(name: string, value: any): void;
    addPluginFile(type: keyof State['pluginsFiles'], file: string): void;
    writeIndexHtml(): Promise<void>;
    writeCommonExports(): Promise<void>;
    writeGlobalImports(): Promise<void>;
    writeClientEntry(): Promise<void>;
    writeServerExports(): Promise<void>;
    writeHookFile(type: 'app' | 'server', projectFiles: string[]): Promise<void>;
}

declare class PluginContext {
    private api;
    pluginName: string;
    constructor(api: Ream, pluginName: string);
    get dev(): boolean;
    get store(): Store;
    get env(): Record<string, string>;
    get constants(): Record<string, string>;
    get config(): MarkRequired<ReamConfig, "env" | "plugins" | "imports">;
    ensureEnv(name: string, defaultValue?: string): string | undefined;
    resolveSrcDir(...args: string[]): string;
    resolveRootDir(...args: string[]): string;
    resolveDotReam(...args: string[]): string;
}

declare type OnFileChangeCallback<This = null> = (this: This, event: 'add' | 'addDir' | 'change' | 'unlink' | 'unlinkDir', filepath: string) => any;
declare type Route = {
    /** A unique id of this route, for example you can use filepath as the id */
    id?: string;
    name: string;
    path: string;
    file: string;
    _nest_key_?: string;
};
declare type NestedRoute = Route & {
    children?: NestedRoute[];
};
declare type TransformServerExports = (this: PluginContext, code: string) => string | Promise<string>;
declare type TransformCommonExports = (this: PluginContext, code: string) => string | Promise<string>;
declare type GetClientRoutes = (this: PluginContext) => Route[] | Promise<Route[]>;
declare type GetApiRoutes = (this: PluginContext) => Route[] | Promise<Route[]>;
declare type ReamPlugin = {
    name: string;
    transformServerExports?: TransformServerExports;
    transformCommonExports?: TransformCommonExports;
    getClientRoutes?: GetClientRoutes;
    getApiRoutes?: GetApiRoutes;
    onFileChange?: OnFileChangeCallback<PluginContext>;
    vite?: (this: PluginContext, config: InlineConfig) => void;
};

declare function normalizePath(path: string): string;

interface Options {
    rootDir?: string;
    dev?: boolean;
    mode?: string;
}

declare type ReamConfig = {
    /**
     * The directory to load `pages`, `api`, `ream-app.js` etc
     * Defaults to root dir (`.`)
     */
    srcDir?: string;
    env?: Record<string, string>;
    plugins?: Array<ReamPlugin>;
    imports?: string[];
    hmr?: {
        host?: string;
        port?: number;
    };
    ssr?: boolean;
    vite?: (viteConfig: UserConfig, opts: {
        dev: boolean;
    }) => void;
    apiRoutes?: (defaultRoutes: Route[]) => Promise<Route[]> | Route[];
    clientRoutes?: (defaultRoutes: Route[]) => Promise<Route[]> | Route[];
};
declare const defineConfig: (config: ReamConfig) => ReamConfig;
declare class Ream {
    rootDir: string;
    isDev: boolean;
    inlineConfig: ReamConfig;
    config: MarkRequired<ReamConfig, 'env' | 'plugins' | 'imports'>;
    configPath?: string;
    viteServer?: ViteDevServer;
    userEnv: Record<string, string>;
    mode: string;
    store: Store;
    plugins: {
        plugin: ReamPlugin;
        context: PluginContext;
    }[];
    constructor(options?: Options, inlineConfig?: ReamConfig);
    resolveRootDir(...args: string[]): string;
    resolveSrcDir(...args: string[]): string;
    resolveDotReam(...args: string[]): string;
    resolveInPackage(pkg: string, target: string): string;
    loadConfig(): Promise<void>;
    get env(): Record<string, string>;
    get constants(): Record<string, string>;
    getPlugins(): {
        plugin: ReamPlugin;
        context: PluginContext;
    }[];
    prepare(): Promise<void>;
    localResolve(name: string): string | undefined;
    localRequire(name: string): any;
    getRequestHandler(): Promise<(_req: http.IncomingMessage, _res: http.ServerResponse) => void>;
    build({ fullyExport, standalone, }: {
        fullyExport?: boolean;
        standalone?: boolean;
    }): Promise<void>;
}

export { GetApiRoutes, GetClientRoutes, NestedRoute, OnFileChangeCallback, Options, PluginContext, Ream, ReamConfig, ReamPlugin, Route, TransformCommonExports, TransformServerExports, defineConfig, normalizePath };
