import * as http from 'http';
import { IncomingMessage, ServerResponse } from 'http';
import { ParsedUrlQuery } from 'querystring';
import { Router, HTTPMethod } from '@egoist/router';
import { ViteDevServer } from 'vite';
import { Render } from './app';

declare type SendData = NodeJS.ReadableStream | object | string | Buffer;

declare type SimpleHandleFunction<TReq = any, TRes = any> = (req: TReq, res: TRes) => void | Promise<void>;
declare type NextHandleFunction<TReq = any, TRes = any> = (req: TReq, res: TRes, next: NextFunction) => void | Promise<void>;
declare type HandleFunction<TReq = any, TRes = any> = SimpleHandleFunction<TReq, TRes> | NextHandleFunction<TReq, TRes>;
declare type NextFunction = (err?: ConnectError) => void;
interface ConnectRequest extends IncomingMessage {
    originalUrl: string;
    path: string;
    query: ParsedUrlQuery;
    params: Record<string, string | string[]>;
}
interface ConnectResponse extends ServerResponse {
}
declare type OnError<TReq, TRes> = (error: ConnectError | string, req: TReq, res: TRes) => void;
declare class ConnectError extends Error {
    code?: number;
    status?: number;
}
declare type Options<TReq, TRes> = {
    onError?: OnError<TReq, TRes>;
    onNoMatch?: SimpleHandleFunction<TReq, TRes>;
};
declare class Connect<TReq extends ConnectRequest, TRes extends ConnectResponse> {
    onError: OnError<TReq, TRes>;
    onNoMatch: SimpleHandleFunction<TReq, TRes>;
    router: Router<HandleFunction<TReq, TRes>>;
    constructor(options?: Options<TReq, TRes>);
    add(method: HTTPMethod, route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]): this;
    use(route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]): this;
    use(handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]): this;
    get: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    head: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    patch: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    options: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    connect: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    delete: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    trace: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    post: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    put: (route: string, handler: HandleFunction<TReq, TRes>, ...handlers: HandleFunction<TReq, TRes>[]) => this;
    handler: (_req: IncomingMessage, _res: ServerResponse) => void;
}

declare type ReamServerHandler = (req: ReamServerRequest, res: ReamServerResponse, next: NextFunction) => void | Promise<void>;
interface ReamServerRequest extends ConnectRequest {
    url: string;
    path: string;
    params: {
        [k: string]: any;
    };
}
interface ReamServerResponse extends ConnectResponse {
    send: (data: SendData) => void;
}
declare type ReamServer = Connect<ReamServerRequest, ReamServerResponse>;

/**
 * Hook: `extendServer`
 */
declare type ExtendServer$1 = (context: {
    server: Connect<ReamServerRequest, ReamServerResponse>;
}) => void | Promise<void>;
declare type GetInitialHTMLContext = {
    head: string;
    main: string;
    scripts: string;
    htmlAttrs: string;
    bodyAttrs: string;
};
/**
 * Hook: `getInitialHTML`
 */
declare type GetInitialHTML = (context: GetInitialHTMLContext) => string | undefined | Promise<string | undefined>;

declare type ApiRoute = {
    path: string;
    load: () => Promise<{
        default: any;
    }>;
};
declare type TransformIndexHTML = (url: string, html: string) => Promise<string>;
interface ServerExports {
    apiRoutes: ApiRoute[];
    serverRender: Render;
    serverHandler?: {
        default: ExtendServer;
    };
    appHandler?: {
        default: ExtendServer;
    };
    enhanceApp: {
        callAsync: (name: string, context: any) => Promise<void>;
    };
    enhanceServer: {
        hasExport: (name: string) => boolean;
        getInitialHTML: GetInitialHTML;
        callAsync: (name: string, context: any) => Promise<void>;
    };
}
declare type CreateServerOptions = {
    viteServer?: ViteDevServer;
    /**
     * Development mode
     * @default {false}
     */
    dev?: boolean;
    /**
     * Path to your project
     */
    cwd?: string;
    devMiddleware?: any;
    fixStacktrace?: (err: Error) => void;
    htmlTemplatePath?: string;
};
declare type ExtendServer = (server: ReamServer, context: {
    serverExports: ServerExports;
    htmlTemplatePath: string;
    transformIndexHTML?: TransformIndexHTML;
    initialState: Record<string, any>;
}) => void | Promise<void>;
declare function createHandler(options: CreateServerOptions): Promise<{
    handler: (_req: http.IncomingMessage, _res: http.ServerResponse) => void;
}>;

export { ApiRoute, Connect, ExtendServer$1 as ExtendServer, GetInitialHTML, ReamServerHandler, ReamServerRequest, ReamServerResponse, ServerExports, TransformIndexHTML, createHandler };
